@@ -1,6 +1,7 @@
 /*
  * op function for ltree and lquery
  * Teodor Sigaev <teodor@stack.net>
+ * $PostgreSQL: pgsql/contrib/ltree/lquery_op.c,v 1.11 2006/10/04 00:29:45 momjian Exp $
  */
 
 #include "ltree.h"
@@ -26,7 +27,7 @@ typedef struct
 }	FieldNot;
 
 static char *
-getlexem(char *start, char *end, int *len)
+getlexeme(char *start, char *end, int *len)
 {
 	char	   *ptr;
 
@@ -54,11 +55,11 @@ bool
 				lenq;
 	bool		isok;
 
-	while ((qn = getlexem(qn, endq, &lenq)) != NULL)
+	while ((qn = getlexeme(qn, endq, &lenq)) != NULL)
 	{
 		tn = t->name;
 		isok = false;
-		while ((tn = getlexem(tn, endt, &lent)) != NULL)
+		while ((tn = getlexeme(tn, endt, &lent)) != NULL)
 		{
 			if (
 				(
@@ -93,7 +94,7 @@ checkLevel(lquery_level * curq, ltree_level * curt)
 	{
 		cmpptr = (curvar->flag & LVAR_INCASE) ? pg_strncasecmp : strncmp;
 
-		if (curvar->flag & LVAR_SUBLEXEM)
+		if (curvar->flag & LVAR_SUBLEXEME)
 		{
 			if (compare_subnode(curt, curvar->name, curvar->len, cmpptr, (curvar->flag & LVAR_ANYEND)))
 				return true;
@@ -328,6 +329,10 @@ lt_q_regex(PG_FUNCTION_ARGS)
 		ereport(ERROR,
 				(errcode(ERRCODE_ARRAY_SUBSCRIPT_ERROR),
 				 errmsg("array must be one-dimensional")));
+	if (ARR_HASNULL(_query))
+		ereport(ERROR,
+				(errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED),
+				 errmsg("array must not contain nulls")));
 
 	while (num > 0)
 	{
