@@ -1,12 +1,13 @@
 /*
  * GiST support for ltree
  * Teodor Sigaev <teodor@stack.net>
+ * $PostgreSQL: pgsql/contrib/ltree/ltree_gist.c,v 1.22 2007/11/16 01:12:24 momjian Exp $
  */
 
 #include "ltree.h"
 #include "access/gist.h"
-#include "access/rtree.h"
 #include "access/nbtree.h"
+#include "access/skey.h"
 #include "utils/array.h"
 #include "crc32.h"
 
@@ -58,7 +59,7 @@ PG_FUNCTION_INFO_V1(ltree_consistent);
 Datum		ltree_consistent(PG_FUNCTION_ARGS);
 
 #define ISEQ(a,b)	( (a)->numlevel == (b)->numlevel && ltree_compare(a,b)==0 )
-#define GETENTRY(vec,pos) ((ltree_gist *) DatumGetPointer(((GISTENTRY *) VARDATA(vec))[(pos)].key))
+#define GETENTRY(vec,pos) ((ltree_gist *) DatumGetPointer((vec)->vector[(pos)].key))
 
 Datum
 ltree_compress(PG_FUNCTION_ARGS)
@@ -70,20 +71,17 @@ ltree_compress(PG_FUNCTION_ARGS)
 	{							/* ltree */
 		ltree_gist *key;
 		ltree	   *val = (ltree *) DatumGetPointer(PG_DETOAST_DATUM(entry->key));
-		int4		len = LTG_HDRSIZE + val->len;
+		int4		len = LTG_HDRSIZE + VARSIZE(val);
 
 		key = (ltree_gist *) palloc(len);
-		key->len = len;
+		SET_VARSIZE(key, len);
 		key->flag = LTG_ONENODE;
-		memcpy((void *) LTG_NODE(key), (void *) val, val->len);
-
-		if (PointerGetDatum(val) != entry->key)
-			pfree(val);
+		memcpy((void *) LTG_NODE(key), (void *) val, VARSIZE(val));
 
 		retval = (GISTENTRY *) palloc(sizeof(GISTENTRY));
 		gistentryinit(*retval, PointerGetDatum(key),
 					  entry->rel, entry->page,
-					  entry->offset, key->len, FALSE);
+					  entry->offset, FALSE);
 	}
 	PG_RETURN_POINTER(retval);
 }
@@ -100,7 +98,7 @@ ltree_decompress(PG_FUNCTION_ARGS)
 
 		gistentryinit(*retval, PointerGetDatum(key),
 					  entry->rel, entry->page,
-					  entry->offset, key->len, FALSE);
+					  entry->offset, FALSE);
 		PG_RETURN_POINTER(retval);
 	}
 	PG_RETURN_POINTER(entry);
@@ -135,13 +133,16 @@ ltree_same(PG_FUNCTION_ARGS)
 
 		*result = true;
 		if (!LTG_ISALLTRUE(a))
-			LOOPBYTE(
-					 if (sa[i] != sb[i])
-					 {
-				*result = false;
-				break;
+		{
+			LOOPBYTE
+			{
+				if (sa[i] != sb[i])
+				{
+					*result = false;
+					break;
+				}
 			}
-		);
+		}
 	}
 
 	PG_RETURN_POINTER(result);
@@ -166,10 +167,9 @@ hashing(BITVECP sign, ltree * t)
 Datum
 ltree_union(PG_FUNCTION_ARGS)
 {
-	bytea	   *entryvec = (bytea *) PG_GETARG_POINTER(0);
+	GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
 	int		   *size = (int *) PG_GETARG_POINTER(1);
 	BITVEC		base;
-	int4		len = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
 	int4		i,
 				j;
 	ltree_gist *result,
@@ -181,7 +181,7 @@ ltree_union(PG_FUNCTION_ARGS)
 	bool		isleqr;
 
 	MemSet((void *) base, 0, sizeof(BITVEC));
-	for (j = 0; j < len; j++)
+	for (j = 0; j < entryvec->n; j++)
 	{
 		cur = GETENTRY(entryvec, j);
 		if (LTG_ISONENODE(cur))
@@ -201,9 +201,8 @@ ltree_union(PG_FUNCTION_ARGS)
 			{
 				BITVECP		sc = LTG_SIGN(cur);
 
-				LOOPBYTE(
-						 ((unsigned char *) base)[i] |= sc[i];
-				);
+				LOOPBYTE
+					((unsigned char *) base)[i] |= sc[i];
 			}
 
 			curtree = LTG_LNODE(cur);
@@ -218,20 +217,21 @@ ltree_union(PG_FUNCTION_ARGS)
 	if (isalltrue == false)
 	{
 		isalltrue = true;
-		LOOPBYTE(
-				 if (((unsigned char *) base)[i] != 0xff)
-				 {
-			isalltrue = false;
-			break;
+		LOOPBYTE
+		{
+			if (((unsigned char *) base)[i] != 0xff)
+			{
+				isalltrue = false;
+				break;
+			}
 		}
-		);
 	}
 
 	isleqr = (left == right || ISEQ(left, right)) ? true : false;
-	*size = LTG_HDRSIZE + ((isalltrue) ? 0 : SIGLEN) + left->len + ((isleqr) ? 0 : right->len);
+	*size = LTG_HDRSIZE + ((isalltrue) ? 0 : SIGLEN) + VARSIZE(left) + ((isleqr) ? 0 : VARSIZE(right));
 
 	result = (ltree_gist *) palloc(*size);
-	result->len = *size;
+	SET_VARSIZE(result, *size);
 	result->flag = 0;
 
 	if (isalltrue)
@@ -239,11 +239,11 @@ ltree_union(PG_FUNCTION_ARGS)
 	else
 		memcpy((void *) LTG_SIGN(result), base, SIGLEN);
 
-	memcpy((void *) LTG_LNODE(result), (void *) left, left->len);
+	memcpy((void *) LTG_LNODE(result), (void *) left, VARSIZE(left));
 	if (isleqr)
 		result->flag |= LTG_NORIGHT;
 	else
-		memcpy((void *) LTG_RNODE(result), (void *) right, right->len);
+		memcpy((void *) LTG_RNODE(result), (void *) right, VARSIZE(right));
 
 	PG_RETURN_POINTER(result);
 }
@@ -260,7 +260,7 @@ ltree_penalty(PG_FUNCTION_ARGS)
 	cmpl = ltree_compare(LTG_GETLNODE(origval), LTG_GETLNODE(newval));
 	cmpr = ltree_compare(LTG_GETRNODE(newval), LTG_GETRNODE(origval));
 
-	*penalty = max(cmpl, 0) + max(cmpr, 0);
+	*penalty = Max(cmpl, 0) + Max(cmpr, 0);
 
 	PG_RETURN_POINTER(penalty);
 }
@@ -278,14 +278,14 @@ treekey_cmp(const void *a, const void *b)
 	return ltree_compare(
 						 ((RIX *) a)->r,
 						 ((RIX *) b)->r
-	);
+		);
 }
 
 
 Datum
 ltree_picksplit(PG_FUNCTION_ARGS)
 {
-	bytea	   *entryvec = (bytea *) PG_GETARG_POINTER(0);
+	GistEntryVector *entryvec = (GistEntryVector *) PG_GETARG_POINTER(0);
 	GIST_SPLITVEC *v = (GIST_SPLITVEC *) PG_GETARG_POINTER(1);
 	OffsetNumber j;
 	int4		i;
@@ -307,7 +307,7 @@ ltree_picksplit(PG_FUNCTION_ARGS)
 
 	memset((void *) ls, 0, sizeof(BITVEC));
 	memset((void *) rs, 0, sizeof(BITVEC));
-	maxoff = ((VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY)) - 1;
+	maxoff = entryvec->n - 1;
 	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
 	v->spl_left = (OffsetNumber *) palloc(nbytes);
 	v->spl_right = (OffsetNumber *) palloc(nbytes);
@@ -346,9 +346,8 @@ ltree_picksplit(PG_FUNCTION_ARGS)
 				{
 					BITVECP		sc = LTG_SIGN(lu);
 
-					LOOPBYTE(
-							 ((unsigned char *) ls)[i] |= sc[i];
-					);
+					LOOPBYTE
+						((unsigned char *) ls)[i] |= sc[i];
 				}
 			}
 		}
@@ -368,9 +367,8 @@ ltree_picksplit(PG_FUNCTION_ARGS)
 				{
 					BITVECP		sc = LTG_SIGN(lu);
 
-					LOOPBYTE(
-							 ((unsigned char *) rs)[i] |= sc[i];
-					);
+					LOOPBYTE
+						((unsigned char *) rs)[i] |= sc[i];
 				}
 			}
 		}
@@ -379,61 +377,62 @@ ltree_picksplit(PG_FUNCTION_ARGS)
 	if (lisat == false)
 	{
 		lisat = true;
-		LOOPBYTE(
-				 if (((unsigned char *) ls)[i] != 0xff)
-				 {
-			lisat = false;
-			break;
+		LOOPBYTE
+		{
+			if (((unsigned char *) ls)[i] != 0xff)
+			{
+				lisat = false;
+				break;
+			}
 		}
-		);
 	}
 
 	if (risat == false)
 	{
 		risat = true;
-		LOOPBYTE(
-				 if (((unsigned char *) rs)[i] != 0xff)
-				 {
-			risat = false;
-			break;
+		LOOPBYTE
+		{
+			if (((unsigned char *) rs)[i] != 0xff)
+			{
+				risat = false;
+				break;
+			}
 		}
-		);
 	}
 
 	lu_l = LTG_GETLNODE(GETENTRY(entryvec, array[FirstOffsetNumber].index));
 	isleqr = (lu_l == lu_r || ISEQ(lu_l, lu_r)) ? true : false;
-	size = LTG_HDRSIZE + ((lisat) ? 0 : SIGLEN) + lu_l->len + ((isleqr) ? 0 : lu_r->len);
+	size = LTG_HDRSIZE + ((lisat) ? 0 : SIGLEN) + VARSIZE(lu_l) + ((isleqr) ? 0 : VARSIZE(lu_r));
 	lu = (ltree_gist *) palloc(size);
-	lu->len = size;
+	SET_VARSIZE(lu, size);
 	lu->flag = 0;
 	if (lisat)
 		lu->flag |= LTG_ALLTRUE;
 	else
 		memcpy((void *) LTG_SIGN(lu), ls, SIGLEN);
-	memcpy((void *) LTG_LNODE(lu), (void *) lu_l, lu_l->len);
+	memcpy((void *) LTG_LNODE(lu), (void *) lu_l, VARSIZE(lu_l));
 	if (isleqr)
 		lu->flag |= LTG_NORIGHT;
 	else
-		memcpy((void *) LTG_RNODE(lu), (void *) lu_r, lu_r->len);
+		memcpy((void *) LTG_RNODE(lu), (void *) lu_r, VARSIZE(lu_r));
 
 
 	ru_l = LTG_GETLNODE(GETENTRY(entryvec, array[1 + ((maxoff - FirstOffsetNumber + 1) / 2)].index));
 	isleqr = (ru_l == ru_r || ISEQ(ru_l, ru_r)) ? true : false;
-	size = LTG_HDRSIZE + ((risat) ? 0 : SIGLEN) + ru_l->len + ((isleqr) ? 0 : ru_r->len);
+	size = LTG_HDRSIZE + ((risat) ? 0 : SIGLEN) + VARSIZE(ru_l) + ((isleqr) ? 0 : VARSIZE(ru_r));
 	ru = (ltree_gist *) palloc(size);
-	ru->len = size;
+	SET_VARSIZE(ru, size);
 	ru->flag = 0;
 	if (risat)
 		ru->flag |= LTG_ALLTRUE;
 	else
 		memcpy((void *) LTG_SIGN(ru), rs, SIGLEN);
-	memcpy((void *) LTG_LNODE(ru), (void *) ru_l, ru_l->len);
+	memcpy((void *) LTG_LNODE(ru), (void *) ru_l, VARSIZE(ru_l));
 	if (isleqr)
 		ru->flag |= LTG_NORIGHT;
 	else
-		memcpy((void *) LTG_RNODE(ru), (void *) ru_r, ru_r->len);
+		memcpy((void *) LTG_RNODE(ru), (void *) ru_r, VARSIZE(ru_r));
 
-	pfree(array);
 	v->spl_ldatum = PointerGetDatum(lu);
 	v->spl_rdatum = PointerGetDatum(ru);
 
@@ -460,29 +459,37 @@ gist_isparent(ltree_gist * key, ltree * query)
 	return false;
 }
 
+static ltree *
+copy_ltree(ltree * src)
+{
+	ltree	   *dst = (ltree *) palloc(VARSIZE(src));
+
+	memcpy(dst, src, VARSIZE(src));
+	return dst;
+}
+
 static bool
 gist_ischild(ltree_gist * key, ltree * query)
 {
-	ltree	   *left = LTG_GETLNODE(key);
-	ltree	   *right = LTG_GETRNODE(key);
-	int4		numlevelL = left->numlevel;
-	int4		numlevelR = right->numlevel;
+	ltree	   *left = copy_ltree(LTG_GETLNODE(key));
+	ltree	   *right = copy_ltree(LTG_GETRNODE(key));
 	bool		res = true;
 
-	if (numlevelL > query->numlevel)
+	if (left->numlevel > query->numlevel)
 		left->numlevel = query->numlevel;
 
 	if (ltree_compare(query, left) < 0)
 		res = false;
 
-	if (numlevelR > query->numlevel)
+	if (right->numlevel > query->numlevel)
 		right->numlevel = query->numlevel;
 
 	if (res && ltree_compare(query, right) > 0)
 		res = false;
 
-	left->numlevel = numlevelL;
-	right->numlevel = numlevelR;
+	pfree(left);
+	pfree(right);
+
 	return res;
 }
 
@@ -538,7 +545,7 @@ gist_tqcmp(ltree * t, lquery * q)
 	while (an > 0 && bn > 0)
 	{
 		bl = LQL_FIRST(ql);
-		if ((res = strncmp(al->name, bl->name, min(al->len, bl->len))) == 0)
+		if ((res = strncmp(al->name, bl->name, Min(al->len, bl->len))) == 0)
 		{
 			if (al->len != bl->len)
 				return al->len - bl->len;
@@ -551,36 +558,22 @@ gist_tqcmp(ltree * t, lquery * q)
 		ql = LQL_NEXT(ql);
 	}
 
-	return t->numlevel - q->firstgood;
+	return Min(t->numlevel, q->firstgood) - q->firstgood;
 }
 
 static bool
 gist_between(ltree_gist * key, lquery * query)
 {
-	ltree	   *left = LTG_GETLNODE(key);
-	ltree	   *right = LTG_GETRNODE(key);
-	int4		numlevelL = left->numlevel;
-	int4		numlevelR = right->numlevel;
-	bool		res = true;
-
 	if (query->firstgood == 0)
 		return true;
 
-	if (numlevelL > query->firstgood)
-		left->numlevel = query->firstgood;
+	if (gist_tqcmp(LTG_GETLNODE(key), query) > 0)
+		return false;
 
-	if (gist_tqcmp(left, query) > 0)
-		res = false;
+	if (gist_tqcmp(LTG_GETRNODE(key), query) < 0)
+		return false;
 
-	if (numlevelR > query->firstgood)
-		right->numlevel = query->firstgood;
-
-	if (res && gist_tqcmp(right, query) < 0)
-		res = false;
-
-	left->numlevel = numlevelL;
-	right->numlevel = numlevelR;
-	return res;
+	return true;
 }
 
 static bool
@@ -603,21 +596,27 @@ gist_qtxt(ltree_gist * key, ltxtquery * query)
 }
 
 static bool
-arrq_cons(ltree_gist *key, ArrayType *_query) {
-        lquery  *query = (lquery *) ARR_DATA_PTR(_query);
-        int     num = ArrayGetNItems(ARR_NDIM(_query), ARR_DIMS(_query));
+arrq_cons(ltree_gist * key, ArrayType *_query)
+{
+	lquery	   *query = (lquery *) ARR_DATA_PTR(_query);
+	int			num = ArrayGetNItems(ARR_NDIM(_query), ARR_DIMS(_query));
 
-        if (ARR_NDIM(_query) != 1)
-			ereport(ERROR,
+	if (ARR_NDIM(_query) != 1)
+		ereport(ERROR,
 				(errcode(ERRCODE_ARRAY_SUBSCRIPT_ERROR),
 				 errmsg("array must be one-dimensional")));
+	if (ARR_HASNULL(_query))
+		ereport(ERROR,
+				(errcode(ERRCODE_NULL_VALUE_NOT_ALLOWED),
+				 errmsg("array must not contain nulls")));
 
-        while (num > 0) {
-		if ( gist_qe(key, query) && gist_between(key, query) ) 
-                        return true;
-                num--;
-                query = NEXTVAL(query);
-        }
+	while (num > 0)
+	{
+		if (gist_qe(key, query) && gist_between(key, query))
+			return true;
+		num--;
+		query = NEXTVAL(query);
+	}
 	return false;
 }
 
@@ -625,52 +624,55 @@ Datum
 ltree_consistent(PG_FUNCTION_ARGS)
 {
 	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
-	char	   *query = (char *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));
+	void	   *query = NULL;
 	ltree_gist *key = (ltree_gist *) DatumGetPointer(entry->key);
 	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
 	bool		res = false;
 
-#ifndef assert_enabled
-#define assert_enabled 0
-#endif
-
 	switch (strategy)
 	{
 		case BTLessStrategyNumber:
+			query = PG_GETARG_LTREE(1);
 			res = (GIST_LEAF(entry)) ?
 				(ltree_compare((ltree *) query, LTG_NODE(key)) > 0)
 				:
 				(ltree_compare((ltree *) query, LTG_GETLNODE(key)) >= 0);
 			break;
 		case BTLessEqualStrategyNumber:
+			query = PG_GETARG_LTREE(1);
 			res = (ltree_compare((ltree *) query, LTG_GETLNODE(key)) >= 0);
 			break;
 		case BTEqualStrategyNumber:
+			query = PG_GETARG_LTREE(1);
 			if (GIST_LEAF(entry))
 				res = (ltree_compare((ltree *) query, LTG_NODE(key)) == 0);
 			else
 				res = (
-				   ltree_compare((ltree *) query, LTG_GETLNODE(key)) >= 0
+					   ltree_compare((ltree *) query, LTG_GETLNODE(key)) >= 0
 					   &&
-				   ltree_compare((ltree *) query, LTG_GETRNODE(key)) <= 0
+					   ltree_compare((ltree *) query, LTG_GETRNODE(key)) <= 0
 					);
 			break;
 		case BTGreaterEqualStrategyNumber:
+			query = PG_GETARG_LTREE(1);
 			res = (ltree_compare((ltree *) query, LTG_GETRNODE(key)) <= 0);
 			break;
 		case BTGreaterStrategyNumber:
+			query = PG_GETARG_LTREE(1);
 			res = (GIST_LEAF(entry)) ?
 				(ltree_compare((ltree *) query, LTG_GETRNODE(key)) < 0)
 				:
 				(ltree_compare((ltree *) query, LTG_GETRNODE(key)) <= 0);
 			break;
 		case 10:
+			query = PG_GETARG_LTREE_COPY(1);
 			res = (GIST_LEAF(entry)) ?
 				inner_isparent((ltree *) query, LTG_NODE(key))
 				:
 				gist_isparent(key, (ltree *) query);
 			break;
 		case 11:
+			query = PG_GETARG_LTREE(1);
 			res = (GIST_LEAF(entry)) ?
 				inner_isparent(LTG_NODE(key), (ltree *) query)
 				:
@@ -678,30 +680,33 @@ ltree_consistent(PG_FUNCTION_ARGS)
 			break;
 		case 12:
 		case 13:
+			query = PG_GETARG_LQUERY(1);
 			if (GIST_LEAF(entry))
 				res = DatumGetBool(DirectFunctionCall2(ltq_regex,
-										  PointerGetDatum(LTG_NODE(key)),
-										PointerGetDatum((lquery *) query)
+											  PointerGetDatum(LTG_NODE(key)),
+											PointerGetDatum((lquery *) query)
 													   ));
 			else
 				res = (gist_qe(key, (lquery *) query) && gist_between(key, (lquery *) query));
 			break;
 		case 14:
 		case 15:
+			query = PG_GETARG_LQUERY(1);
 			if (GIST_LEAF(entry))
 				res = DatumGetBool(DirectFunctionCall2(ltxtq_exec,
-										  PointerGetDatum(LTG_NODE(key)),
-										PointerGetDatum((lquery *) query)
+											  PointerGetDatum(LTG_NODE(key)),
+											PointerGetDatum((lquery *) query)
 													   ));
 			else
 				res = gist_qtxt(key, (ltxtquery *) query);
 			break;
 		case 16:
 		case 17:
+			query = DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));
 			if (GIST_LEAF(entry))
 				res = DatumGetBool(DirectFunctionCall2(lt_q_regex,
-										  PointerGetDatum(LTG_NODE(key)),
-										PointerGetDatum((ArrayType *) query)
+											  PointerGetDatum(LTG_NODE(key)),
+										 PointerGetDatum((ArrayType *) query)
 													   ));
 			else
 				res = arrq_cons(key, (ArrayType *) query);
@@ -710,5 +715,7 @@ ltree_consistent(PG_FUNCTION_ARGS)
 			/* internal error */
 			elog(ERROR, "unrecognized StrategyNumber: %d", strategy);
 	}
+
+	PG_FREE_IF_COPY(query, 1);
 	PG_RETURN_BOOL(res);
 }
