@@ -7,7 +7,7 @@
  * And contributors:
  * Nabil Sayegh <postgresql@e-trolley.de>
  *
- * Copyright (c) 2002-2006, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2008, PostgreSQL Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose, without fee, and without a written agreement
@@ -44,9 +44,9 @@
 
 PG_MODULE_MAGIC;
 
-static int	load_categories_hash(char *cats_sql, MemoryContext per_query_ctx);
+static HTAB *load_categories_hash(char *cats_sql, MemoryContext per_query_ctx);
 static Tuplestorestate *get_crosstab_tuplestore(char *sql,
-						int num_categories,
+						HTAB *crosstab_hash,
 						TupleDesc tupdesc,
 						MemoryContext per_query_ctx);
 static void validateConnectbyTupleDesc(TupleDesc tupdesc, bool show_branch, bool show_serial);
@@ -121,26 +121,23 @@ typedef struct
 /* sign, 10 digits, '\0' */
 #define INT32_STRLEN	12
 
-/* hash table support */
-static HTAB *crosstab_HashTable;
-
-/* The information we cache about loaded procedures */
+/* stored info for a crosstab category */
 typedef struct crosstab_cat_desc
 {
-	char	   *catname;
+	char	   *catname;		/* full category name */
 	int			attidx;			/* zero based */
 }	crosstab_cat_desc;
 
 #define MAX_CATNAME_LEN			NAMEDATALEN
 #define INIT_CATS				64
 
-#define crosstab_HashTableLookup(CATNAME, CATDESC) \
+#define crosstab_HashTableLookup(HASHTAB, CATNAME, CATDESC) \
 do { \
 	crosstab_HashEnt *hentry; char key[MAX_CATNAME_LEN]; \
 	\
 	MemSet(key, 0, MAX_CATNAME_LEN); \
 	snprintf(key, MAX_CATNAME_LEN - 1, "%s", CATNAME); \
-	hentry = (crosstab_HashEnt*) hash_search(crosstab_HashTable, \
+	hentry = (crosstab_HashEnt*) hash_search(HASHTAB, \
 										 key, HASH_FIND, NULL); \
 	if (hentry) \
 		CATDESC = hentry->catdesc; \
@@ -148,13 +145,13 @@ do { \
 		CATDESC = NULL; \
 } while(0)
 
-#define crosstab_HashTableInsert(CATDESC) \
+#define crosstab_HashTableInsert(HASHTAB, CATDESC) \
 do { \
 	crosstab_HashEnt *hentry; bool found; char key[MAX_CATNAME_LEN]; \
 	\
 	MemSet(key, 0, MAX_CATNAME_LEN); \
 	snprintf(key, MAX_CATNAME_LEN - 1, "%s", CATDESC->catname); \
-	hentry = (crosstab_HashEnt*) hash_search(crosstab_HashTable, \
+	hentry = (crosstab_HashEnt*) hash_search(HASHTAB, \
 										 key, HASH_ENTER, &found); \
 	if (found) \
 		ereport(ERROR, \
@@ -549,8 +546,8 @@ crosstab(PG_FUNCTION_ARGS)
 					xpstrdup(values[0], rowid);
 
 					/*
-					 * Check to see if the rowid is the same as that of the last
-					 * tuple sent -- if so, skip this tuple entirely
+					 * Check to see if the rowid is the same as that of the
+					 * last tuple sent -- if so, skip this tuple entirely
 					 */
 					if (!firstpass && xstreq(lastrowid, rowid))
 					{
@@ -560,8 +557,8 @@ crosstab(PG_FUNCTION_ARGS)
 				}
 
 				/*
-				 * If rowid hasn't changed on us, continue building the
-				 * ouput tuple.
+				 * If rowid hasn't changed on us, continue building the ouput
+				 * tuple.
 				 */
 				if (xstreq(rowid, values[0]))
 				{
@@ -701,7 +698,7 @@ crosstab_hash(PG_FUNCTION_ARGS)
 	TupleDesc	tupdesc;
 	MemoryContext per_query_ctx;
 	MemoryContext oldcontext;
-	int			num_categories;
+	HTAB	   *crosstab_hash;
 
 	/* check to see if caller supports us returning a tuplestore */
 	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
@@ -734,14 +731,14 @@ crosstab_hash(PG_FUNCTION_ARGS)
 						"crosstab function are not compatible")));
 
 	/* load up the categories hash table */
-	num_categories = load_categories_hash(cats_sql, per_query_ctx);
+	crosstab_hash = load_categories_hash(cats_sql, per_query_ctx);
 
 	/* let the caller know we're sending back a tuplestore */
 	rsinfo->returnMode = SFRM_Materialize;
 
 	/* now go build it */
 	rsinfo->setResult = get_crosstab_tuplestore(sql,
-												num_categories,
+												crosstab_hash,
 												tupdesc,
 												per_query_ctx);
 
@@ -761,24 +758,29 @@ crosstab_hash(PG_FUNCTION_ARGS)
 /*
  * load up the categories hash table
  */
-static int
+static HTAB *
 load_categories_hash(char *cats_sql, MemoryContext per_query_ctx)
 {
+	HTAB	   *crosstab_hash;
 	HASHCTL		ctl;
 	int			ret;
 	int			proc;
 	MemoryContext SPIcontext;
-	int			num_categories = 0;
 
 	/* initialize the category hash table */
+	MemSet(&ctl, 0, sizeof(ctl));
 	ctl.keysize = MAX_CATNAME_LEN;
 	ctl.entrysize = sizeof(crosstab_HashEnt);
+	ctl.hcxt = per_query_ctx;
 
 	/*
 	 * use INIT_CATS, defined above as a guess of how many hash table entries
 	 * to create, initially
 	 */
-	crosstab_HashTable = hash_create("crosstab hash", INIT_CATS, &ctl, HASH_ELEM);
+	crosstab_hash = hash_create("crosstab hash",
+								INIT_CATS,
+								&ctl,
+								HASH_ELEM | HASH_CONTEXT);
 
 	/* Connect to SPI manager */
 	if ((ret = SPI_connect()) < 0)
@@ -787,7 +789,7 @@ load_categories_hash(char *cats_sql, MemoryContext per_query_ctx)
 
 	/* Retrieve the category name rows */
 	ret = SPI_execute(cats_sql, true, 0);
-	num_categories = proc = SPI_processed;
+	proc = SPI_processed;
 
 	/* Check for qualifying tuples */
 	if ((ret == SPI_OK_SELECT) && (proc > 0))
@@ -825,7 +827,7 @@ load_categories_hash(char *cats_sql, MemoryContext per_query_ctx)
 			catdesc->attidx = i;
 
 			/* Add the proc description block to the hashtable */
-			crosstab_HashTableInsert(catdesc);
+			crosstab_HashTableInsert(crosstab_hash, catdesc);
 
 			MemoryContextSwitchTo(SPIcontext);
 		}
@@ -835,7 +837,7 @@ load_categories_hash(char *cats_sql, MemoryContext per_query_ctx)
 		/* internal error */
 		elog(ERROR, "load_categories_hash: SPI_finish() failed");
 
-	return num_categories;
+	return crosstab_hash;
 }
 
 /*
@@ -843,16 +845,18 @@ load_categories_hash(char *cats_sql, MemoryContext per_query_ctx)
  */
 static Tuplestorestate *
 get_crosstab_tuplestore(char *sql,
-						int num_categories,
+						HTAB *crosstab_hash,
 						TupleDesc tupdesc,
 						MemoryContext per_query_ctx)
 {
 	Tuplestorestate *tupstore;
+	int			num_categories = hash_get_num_entries(crosstab_hash);
 	AttInMetadata *attinmeta = TupleDescGetAttInMetadata(tupdesc);
 	char	  **values;
 	HeapTuple	tuple;
 	int			ret;
 	int			proc;
+	MemoryContext SPIcontext;
 
 	/* initialize our tuplestore */
 	tupstore = tuplestore_begin_heap(true, false, work_mem);
@@ -952,7 +956,10 @@ get_crosstab_tuplestore(char *sql,
 					/* rowid changed, flush the previous output row */
 					tuple = BuildTupleFromCStrings(attinmeta, values);
 
+					/* switch to appropriate context while storing the tuple */
+					SPIcontext = MemoryContextSwitchTo(per_query_ctx);
 					tuplestore_puttuple(tupstore, tuple);
+					MemoryContextSwitchTo(SPIcontext);
 
 					for (j = 0; j < result_ncols; j++)
 						xpfree(values[j]);
@@ -971,7 +978,7 @@ get_crosstab_tuplestore(char *sql,
 
 			if (catname != NULL)
 			{
-				crosstab_HashTableLookup(catname, catdesc);
+				crosstab_HashTableLookup(crosstab_hash, catname, catdesc);
 
 				if (catdesc)
 					values[catdesc->attidx + ncols - 2] =
@@ -985,7 +992,10 @@ get_crosstab_tuplestore(char *sql,
 		/* flush the last output row */
 		tuple = BuildTupleFromCStrings(attinmeta, values);
 
+		/* switch to appropriate context while storing the tuple */
+		SPIcontext = MemoryContextSwitchTo(per_query_ctx);
 		tuplestore_puttuple(tupstore, tuple);
+		MemoryContextSwitchTo(SPIcontext);
 	}
 
 	if (SPI_finish() != SPI_OK_FINISH)
@@ -1267,6 +1277,7 @@ build_tuplestore_recursively(char *key_fld,
 							 Tuplestorestate *tupstore)
 {
 	TupleDesc	tupdesc = attinmeta->tupdesc;
+	MemoryContext oldcontext;
 	int			ret;
 	int			proc;
 	int			serial_column;
@@ -1344,9 +1355,15 @@ build_tuplestore_recursively(char *key_fld,
 		/* construct the tuple */
 		tuple = BuildTupleFromCStrings(attinmeta, values);
 
+		/* switch to long lived context while storing the tuple */
+		oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
 		/* now store it */
 		tuplestore_puttuple(tupstore, tuple);
 
+		/* now reset the context */
+		MemoryContextSwitchTo(oldcontext);
+
 		/* increment level */
 		level++;
 	}
@@ -1383,15 +1400,12 @@ build_tuplestore_recursively(char *key_fld,
 								   "incompatible.")));
 		}
 
+		initStringInfo(&branchstr);
+		initStringInfo(&chk_branchstr);
+		initStringInfo(&chk_current_key);
+
 		for (i = 0; i < proc; i++)
 		{
-			/* start a new branch */
-			initStringInfo(&branchstr);
-
-			/* need these to check for recursion */
-			initStringInfo(&chk_branchstr);
-			initStringInfo(&chk_current_key);
-
 			/* initialize branch for this pass */
 			appendStringInfo(&branchstr, "%s", branch);
 			appendStringInfo(&chk_branchstr, "%s%s%s", branch_delim, branch, branch_delim);
@@ -1435,9 +1449,15 @@ build_tuplestore_recursively(char *key_fld,
 			xpfree(current_key);
 			xpfree(current_key_parent);
 
+			/* switch to long lived context while storing the tuple */
+			oldcontext = MemoryContextSwitchTo(per_query_ctx);
+
 			/* store the tuple for later use */
 			tuplestore_puttuple(tupstore, tuple);
 
+			/* now reset the context */
+			MemoryContextSwitchTo(oldcontext);
+
 			heap_freetuple(tuple);
 
 			/* recurse using current_key_parent as the new start_with */
@@ -1458,10 +1478,14 @@ build_tuplestore_recursively(char *key_fld,
 													tupstore);
 
 			/* reset branch for next pass */
-			xpfree(branchstr.data);
-			xpfree(chk_branchstr.data);
-			xpfree(chk_current_key.data);
+			resetStringInfo(&branchstr);
+			resetStringInfo(&chk_branchstr);
+			resetStringInfo(&chk_current_key);
 		}
+
+		xpfree(branchstr.data);
+		xpfree(chk_branchstr.data);
+		xpfree(chk_current_key.data);
 	}
 
 	return tupstore;
