@@ -3,12 +3,12 @@
  * nodeLimit.c
  *	  Routines to handle limiting of query results where appropriate
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/executor/nodeLimit.c,v 1.17 2003/08/04 02:39:59 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/executor/nodeLimit.c,v 1.33 2008/01/01 19:45:49 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -38,7 +38,6 @@ TupleTableSlot *				/* return: a tuple or NULL */
 ExecLimit(LimitState *node)
 {
 	ScanDirection direction;
-	TupleTableSlot *resultTupleSlot;
 	TupleTableSlot *slot;
 	PlanState  *outerPlan;
 
@@ -47,7 +46,6 @@ ExecLimit(LimitState *node)
 	 */
 	direction = node->ps.state->es_direction;
 	outerPlan = outerPlanState(node);
-	resultTupleSlot = node->ps.ps_ResultTupleSlot;
 
 	/*
 	 * The main logic is a simple state machine.
@@ -56,20 +54,24 @@ ExecLimit(LimitState *node)
 	{
 		case LIMIT_INITIAL:
 
+			/*
+			 * First call for this node, so compute limit/offset. (We can't do
+			 * this any earlier, because parameters from upper nodes will not
+			 * be set during ExecInitLimit.)  This also sets position = 0 and
+			 * changes the state to LIMIT_RESCAN.
+			 */
+			recompute_limits(node);
+
+			/* FALL THRU */
+
+		case LIMIT_RESCAN:
+
 			/*
 			 * If backwards scan, just return NULL without changing state.
 			 */
 			if (!ScanDirectionIsForward(direction))
 				return NULL;
 
-			/*
-			 * First call for this scan, so compute limit/offset. (We
-			 * can't do this any earlier, because parameters from upper
-			 * nodes may not be set until now.)  This also sets position =
-			 * 0.
-			 */
-			recompute_limits(node);
-
 			/*
 			 * Check for empty window; if so, treat like empty subplan.
 			 */
@@ -88,8 +90,8 @@ ExecLimit(LimitState *node)
 				if (TupIsNull(slot))
 				{
 					/*
-					 * The subplan returns too few tuples for us to
-					 * produce any output at all.
+					 * The subplan returns too few tuples for us to produce
+					 * any output at all.
 					 */
 					node->lstate = LIMIT_EMPTY;
 					return NULL;
@@ -117,11 +119,10 @@ ExecLimit(LimitState *node)
 			if (ScanDirectionIsForward(direction))
 			{
 				/*
-				 * Forwards scan, so check for stepping off end of window.
-				 * If we are at the end of the window, return NULL without
-				 * advancing the subplan or the position variable; but
-				 * change the state machine state to record having done
-				 * so.
+				 * Forwards scan, so check for stepping off end of window. If
+				 * we are at the end of the window, return NULL without
+				 * advancing the subplan or the position variable; but change
+				 * the state machine state to record having done so.
 				 */
 				if (!node->noCount &&
 					node->position >= node->offset + node->count)
@@ -145,9 +146,8 @@ ExecLimit(LimitState *node)
 			else
 			{
 				/*
-				 * Backwards scan, so check for stepping off start of
-				 * window. As above, change only state-machine status if
-				 * so.
+				 * Backwards scan, so check for stepping off start of window.
+				 * As above, change only state-machine status if so.
 				 */
 				if (node->position <= node->offset + 1)
 				{
@@ -171,9 +171,8 @@ ExecLimit(LimitState *node)
 				return NULL;
 
 			/*
-			 * Backing up from subplan EOF, so re-fetch previous tuple;
-			 * there should be one!  Note previous tuple must be in
-			 * window.
+			 * Backing up from subplan EOF, so re-fetch previous tuple; there
+			 * should be one!  Note previous tuple must be in window.
 			 */
 			slot = ExecProcNode(outerPlan);
 			if (TupIsNull(slot))
@@ -219,16 +218,11 @@ ExecLimit(LimitState *node)
 	/* Return the current tuple */
 	Assert(!TupIsNull(slot));
 
-	ExecStoreTuple(slot->val,
-				   resultTupleSlot,
-				   InvalidBuffer,
-				   false);		/* tuple does not belong to slot */
-
-	return resultTupleSlot;
+	return slot;
 }
 
 /*
- * Evaluate the limit/offset expressions --- done at start of each scan.
+ * Evaluate the limit/offset expressions --- done at startup or rescan.
  *
  * This is also a handy place to reset the current-position state info.
  */
@@ -236,20 +230,24 @@ static void
 recompute_limits(LimitState *node)
 {
 	ExprContext *econtext = node->ps.ps_ExprContext;
+	Datum		val;
 	bool		isNull;
 
 	if (node->limitOffset)
 	{
-		node->offset =
-			DatumGetInt32(ExecEvalExprSwitchContext(node->limitOffset,
-													econtext,
-													&isNull,
-													NULL));
+		val = ExecEvalExprSwitchContext(node->limitOffset,
+										econtext,
+										&isNull,
+										NULL);
 		/* Interpret NULL offset as no offset */
 		if (isNull)
 			node->offset = 0;
-		else if (node->offset < 0)
-			node->offset = 0;
+		else
+		{
+			node->offset = DatumGetInt64(val);
+			if (node->offset < 0)
+				node->offset = 0;
+		}
 	}
 	else
 	{
@@ -259,17 +257,23 @@ recompute_limits(LimitState *node)
 
 	if (node->limitCount)
 	{
-		node->noCount = false;
-		node->count =
-			DatumGetInt32(ExecEvalExprSwitchContext(node->limitCount,
-													econtext,
-													&isNull,
-													NULL));
+		val = ExecEvalExprSwitchContext(node->limitCount,
+										econtext,
+										&isNull,
+										NULL);
 		/* Interpret NULL count as no count (LIMIT ALL) */
 		if (isNull)
-			node->noCount = true;
-		else if (node->count < 0)
+		{
 			node->count = 0;
+			node->noCount = true;
+		}
+		else
+		{
+			node->count = DatumGetInt64(val);
+			if (node->count < 0)
+				node->count = 0;
+			node->noCount = false;
+		}
 	}
 	else
 	{
@@ -281,6 +285,41 @@ recompute_limits(LimitState *node)
 	/* Reset position to start-of-scan */
 	node->position = 0;
 	node->subSlot = NULL;
+
+	/* Set state-machine state */
+	node->lstate = LIMIT_RESCAN;
+
+	/*
+	 * If we have a COUNT, and our input is a Sort node, notify it that it can
+	 * use bounded sort.
+	 *
+	 * This is a bit of a kluge, but we don't have any more-abstract way of
+	 * communicating between the two nodes; and it doesn't seem worth trying
+	 * to invent one without some more examples of special communication
+	 * needs.
+	 *
+	 * Note: it is the responsibility of nodeSort.c to react properly to
+	 * changes of these parameters.  If we ever do redesign this, it'd be a
+	 * good idea to integrate this signaling with the parameter-change
+	 * mechanism.
+	 */
+	if (IsA(outerPlanState(node), SortState))
+	{
+		SortState  *sortState = (SortState *) outerPlanState(node);
+		int64		tuples_needed = node->count + node->offset;
+
+		/* negative test checks for overflow */
+		if (node->noCount || tuples_needed < 0)
+		{
+			/* make sure flag gets reset if needed upon rescan */
+			sortState->bounded = false;
+		}
+		else
+		{
+			sortState->bounded = true;
+			sortState->bound = tuples_needed;
+		}
+	}
 }
 
 /* ----------------------------------------------------------------
@@ -291,11 +330,14 @@ recompute_limits(LimitState *node)
  * ----------------------------------------------------------------
  */
 LimitState *
-ExecInitLimit(Limit *node, EState *estate)
+ExecInitLimit(Limit *node, EState *estate, int eflags)
 {
 	LimitState *limitstate;
 	Plan	   *outerPlan;
 
+	/* check for unsupported flags */
+	Assert(!(eflags & EXEC_FLAG_MARK));
+
 	/*
 	 * create state structure
 	 */
@@ -324,7 +366,7 @@ ExecInitLimit(Limit *node, EState *estate)
 #define LIMIT_NSLOTS 1
 
 	/*
-	 * Tuple table initialization
+	 * Tuple table initialization (XXX not actually used...)
 	 */
 	ExecInitResultTupleSlot(estate, &limitstate->ps);
 
@@ -332,13 +374,13 @@ ExecInitLimit(Limit *node, EState *estate)
 	 * then initialize outer plan
 	 */
 	outerPlan = outerPlan(node);
-	outerPlanState(limitstate) = ExecInitNode(outerPlan, estate);
+	outerPlanState(limitstate) = ExecInitNode(outerPlan, estate, eflags);
 
 	/*
-	 * limit nodes do no projections, so initialize projection info for
-	 * this node appropriately
+	 * limit nodes do no projections, so initialize projection info for this
+	 * node appropriately
 	 */
-	ExecAssignResultTypeFromOuterPlan(&limitstate->ps);
+	ExecAssignResultTypeFromTL(&limitstate->ps);
 	limitstate->ps.ps_ProjInfo = NULL;
 
 	return limitstate;
@@ -363,10 +405,6 @@ void
 ExecEndLimit(LimitState *node)
 {
 	ExecFreeExprContext(&node->ps);
-
-	/* clean up tuple table */
-	ExecClearTuple(node->ps.ps_ResultTupleSlot);
-
 	ExecEndNode(outerPlanState(node));
 }
 
@@ -374,10 +412,12 @@ ExecEndLimit(LimitState *node)
 void
 ExecReScanLimit(LimitState *node, ExprContext *exprCtxt)
 {
-	/* resetting lstate will force offset/limit recalculation */
-	node->lstate = LIMIT_INITIAL;
-
-	ExecClearTuple(node->ps.ps_ResultTupleSlot);
+	/*
+	 * Recompute limit/offset in case parameters changed, and reset the state
+	 * machine.  We must do this before rescanning our child node, in case
+	 * it's a Sort that we are passing the parameters down to.
+	 */
+	recompute_limits(node);
 
 	/*
 	 * if chgParam of subnode is not null then plan will be re-scanned by
